
def proxyto(target, source, allowed_specials=[]):
  """Proxy all attributes to another object.

  Copies all attributes from `source` to `target`, excepting "special"
  methods (those with leading and trailing double-underscores).
  `allowed_specials` allows you to whitelist selected special methods.

  Parameters
  ----------
  target : object
      object to copy attributes to
  source : object
      object to copy attributes from
  allowed_specials : [str], optional
      names of special methods to copy.

  Returns
  -------
  target : object
  """

  # set all non-hidden methods
  for k in dir(source):
    k_is_special = k.startswith("__") and k.endswith("__")
    copy_k       = not k_is_special or k in allowed_specials
    if not hasattr(target, k) and copy_k:
      setattr(target, k, getattr(source, k))

  return target



from threading import Thread
from concurrent.futures import Executor, ThreadPoolExecutor, Future


class UnboundedThreadPoolExecutor(Executor):
  """
  A thread pool with an infinite number of threads.

  This interface conforms to the typical `concurrent.futures.Executor`
  interface, but doesn't limit the user to a finite number of threads. In
  normal situations, this is undesirable -- too many threads, and your program
  will spend more time switching contexts than actually working!

  On the other hand, if you patch the `thread` module with `gevent`, spawning
  tens of thousands of threads is totally OK. This is where this executor comes
  in.

  Parameters
  ----------
  max_workers: None or int, optional
      Number of worker threads. If None, a new thread is created every time a
      new task is submitted. If an integer, this executor acts exactly like a
      normal `concurrent.futures.ThreadPoolExecutor`.
  """

  def __init__(self, max_workers=None):
    self._open       = True
    self.max_workers = max_workers

    if max_workers is None:
      self._pool = None
    else:
      # finite number of workers? just use a ThreadPoolExecutor
      self._pool = ThreadPoolExecutor(max_workerss)

  def submit(self, fn, *args, **kwargs):
    """Submit a new task to be executed asynchronously.

    If `self.max_workers` is an integer, then the behavior of this function
    will be identical to that of `concurrent.futures.ThreadPoolExecutor`.
    However, if it is None, then a new daemonized thread will be constructed
    and started.

    Parameters
    ----------
    fn : function-like
        function (or callable object) to execute asynchronously.
    args : list
        positional arguments to pass to `fn`.
    kwargs: dict
        keyword arguments to pass to `fn`.

    Returns
    -------
    future : concurrent.futures.Future
        Container for future result or exception
    """
    if not hasattr(fn, '__call__'):
      raise ValueError("`fn` argument isn't callable!")

    if not self._open:
      raise RuntimeError(
        "UnboundedThreadPoolExecutor is already closed. You cannot submit "
        "any more tasks."
      )

    if self._pool is not None:
      return self._pool.submit(fn, *args, **kwargs)
    else:
      # future to contain output/exception
      future = Future()
      future.set_running_or_notify_cancel()

      # run function, save output
      def populate():
        try:
          future.set_result(fn(*args, **kwargs))
        except Exception as e:
          future.set_exception(e)

      # create/start a new thread not bound to any threadpool
      thread = Thread(target=populate)
      thread.daemon = True
      thread.start()

      return future

  def shutdown(self, wait=True):
    """Shutdown this thread pool, preventing future tasks from being enqueued.

    Parameters
    ----------
    wait : bool
        Wait for all running threads to finish. Only used if this pool was
        initialized with a fixed number of workers.
    """
    self._open = False
    if self._pool is not None:
      self._pool.shutdown(wait=wait)



import sys

import joblib


class MiraiError(Exception):
  """Base class for all exceptions raise by Promises"""
  pass


class AlreadyResolvedError(MiraiError):
  """
  Exception thrown when attempting to set the value or exception of a Promise
  that has already had its value or exception set.
  """
  pass


class ShadowException(MiraiError):
  """
  An exception that's never used directly. In particular, a ShadowException is
  used to wrap a thrown exception and allow it to store the traceback context
  as a string so that users of Promise will have some idea what's going on.

  Parameters
  ----------
  exception : Exception
      Other exception to wrap
  context : string
      Formatted string containing the traceback that caused `other` to be thrown
  """

  def __init__(self, exception, context):
    self.exception = exception
    self.context   = context

  def __unicode__(self):
    return u"{cls}: {msg}\n{ctx}".format(
      cls = self.__class__.__name__,
      msg = self.message,
      ctx = self.context,
    )

  def __str__(self):
    return unicode(self)

  def __repr__(self):
    return u"{cls}({msg})".format(
      cls = self.__class__.__name__,
      msg = self.message,
    )

  def __getattr__(self, key):
    return getattr(self.exception, key)

  @staticmethod
  def build(exception, context):
    """
    Construct a child class of a thrown exception and ShadowException, and
    instantiate it.
    """
    cls = exception.__class__
    t = type(
      "Mirai" + cls.__name__,
      (ShadowException, cls),
      {}
    )
    return t(exception, context)


class SafeFunction(object):
  """
  A function-like object that catches all errors and adds their execution stack
  description to their message.
  """

  def __init__(self, f):
    self.f = f

  def __call__(self, *args, **kwargs):
    try:
      return self.f(*args, **kwargs)
    except Exception as e:
      if isinstance(e, ShadowException):
        raise e
      else:
        e_type, e_value, e_tb = sys.exc_info()

        # turn stack into a string
        text = joblib.format_stack.format_exc(e_type, e_value, e_tb, context=10, tb_offset=1)

        # manually delete e_tb (failing to do so will cause a memory leak. See
        # documentation for sys.exc_info())
        del e_tb

        if isinstance(e, MiraiError):
          # You'll get an invalid MRO if you inherit from MiraiError _and_
          # ShadowException (as the latter is already a child of MiraiError).
          # In this case, there's no need to construct a new type.
          raise ShadowException(e, text)
        else:
          # construct a new exception instance that's of the same class as the one
          # thrown, but also with additional context.
          raise ShadowException.build(e, text)



from concurrent.futures import wait
import unittest

from mirai import Promise
from mirai.pool import *
from mirai.tests.test_futures import (
    PromiseConstructorTests,
    PromiseBasicTests,
    PromiseCallbackTests,
    PromiseMapTests,
    PromiseMiscellaneousTests,
    PromiseAlternativeNamesTests,
    PromiseMergingTests,
)


class UnboundedThreadPoolExecutorTests(unittest.TestCase):

  def setUp(self):
    self.pool = UnboundedThreadPoolExecutor()

  def test_submit(self):

    def foo(a, b, c):
      return a + b * c

    futures = [self.pool.submit(foo, i, i+1, c=i+2) for i in range(100)]
    wait(futures)
    results = [f.result() for f in futures]

    self.assertEqual(
      [foo(i, i+1, c=i+2) for i in range(100)],
      results,
    )

  def test_shutdown(self):
    self.pool.shutdown()
    self.assertRaises(RuntimeError, self.pool.submit, lambda: 1)


class UnboundedThreadPoolExecutorPromiseTests(
    PromiseConstructorTests,
    PromiseBasicTests,
    PromiseCallbackTests,
    PromiseMapTests,
    PromiseMiscellaneousTests,
    PromiseAlternativeNamesTests,
    PromiseMergingTests,

    unittest.TestCase
  ):

  def setUp(self):
    self.pool = UnboundedThreadPoolExecutor()
    Promise.executor(self.pool)



from concurrent.futures import ThreadPoolExecutor
import unittest

from mirai import *


class ExampleException(Exception):
  pass


class PromiseConstructorTests(object):
  """Constructing new Promises without any dependencies."""

  def test_value(self):
    self.assertEqual(Promise.value(1).get(0.5), 1)

  def test_exception(self):
    self.assertRaises(ExampleException, Promise.exception(ExampleException()).get, 0.05)

  def test_wait(self):
    self.assertIsNone(Promise.wait(0.05).get(0.5), None)

  def test_eval(self):

    def foo(a, b):
      return a+b

    # value matches
    self.assertEqual(3, Promise.eval(foo, 1, b=2).get(0))

    # exceptions subclass appropriately
    def bar():
      raise NotImplementedError("Uh oh...")

    self.assertRaises(NotImplementedError, Promise.eval(bar).get, 0)
    self.assertRaises(MiraiError, Promise.eval(bar).get, 0)

  def test_call(self):

    def foo(a, b):
      return a+b

    # value matches
    self.assertEqual(3, Promise.call(foo, 1, b=2).get(0.5))

    # exceptions subclass appropriately
    def bar():
      raise NotImplementedError("Uh oh...")

    self.assertRaises(NotImplementedError, Promise.call(bar).get, 0.5)
    self.assertRaises(MiraiError, Promise.call(bar).get, 0.5)


class PromiseBasicTests(object):
  """Core Promise functionality: setting/getting state"""

  def test_setvalue(self):
    o    = object()
    fut1 = Promise()
    fut1.setvalue(o)

    self.assertEqual(fut1.get(0.5), o)

  def test_setvalue_twice(self):
    fut1 = Promise()
    fut1.setvalue(1)

    self.assertRaises(MiraiError, fut1.setvalue, 2)

    fut2 = Promise()
    fut2.setexception(1)

    self.assertRaises(MiraiError, fut2.setvalue, 2)

  def test_setexception(self):
    e    = ExampleException()
    fut1 = Promise()
    fut1.setexception(e)

    self.assertRaises(ExampleException, fut1.get, 0.05)

  def test_setexception_twice(self):
    e    = ExampleException()
    fut1 = Promise()
    fut1.setexception(e)

    self.assertRaises(MiraiError, fut1.setexception, e)

    fut1 = Promise()
    fut1.setvalue(1)

    self.assertRaises(MiraiError, fut1.setexception, e)

  def test_get_success(self):
    fut = Promise.value(1)

    self.assertEqual(fut.get(0.5), 1)

  def test_get_exception(self):
    class VerySpecificException(ExampleException): pass
    fut = Promise.exception(VerySpecificException())

    # raises the right classs
    self.assertRaises(VerySpecificException, fut.get, 0.05)

  def test_isdefined(self):
    self.assertFalse(Promise().isdefined())
    self.assertTrue(Promise.value(1).isdefined())
    self.assertTrue(Promise.exception(ExampleException()).isdefined())

  def test_isfailure(self):
    self.assertIsNone(Promise().isfailure())
    self.assertTrue(Promise.exception(ExampleException()).isfailure())
    self.assertFalse(Promise.value(1).isfailure())

  def test_issuccess(self):
    self.assertIsNone(Promise().issuccess())
    self.assertFalse(Promise.exception(ExampleException()).issuccess())
    self.assertTrue(Promise.value(1).issuccess())

  def test_proxyto(self):
    fut1 = Promise()
    fut2 = Promise.wait(0.05).map(lambda v: 1).proxyto(fut1)

    self.assertEqual(fut1.get(0.1), 1)

    fut1 = Promise()
    fut2 = Promise.wait(0.05) \
        .flatmap(lambda v: Promise.exception(MiraiError())) \
        .proxyto(fut1)

    self.assertRaises(MiraiError, fut1.get, 0.1)


class PromiseCallbackTests(object):
  """Tests for settings callbacks"""

  def test_onsuccess_success(self):
    fut1 = Promise()
    fut2 = Promise.value(1).onsuccess(lambda v: fut1.setvalue(v))

    self.assertEqual(fut1.get(0.5), 1)

  def test_onsuccess_failure(self):
    fut1 = Promise()
    fut2 = Promise.exception(ExampleException()).onsuccess(lambda v: fut1.setvalue(v))

    Promise.join([fut2])

    self.assertRaises(TimeoutError, fut1.within(0).get, 0.05)

  def test_onfailure_exception(self):
    def uhoh(e):
      raise ExampleException("uh oh!")

    fut1 = Promise.value(0).onfailure(uhoh)

    self.assertEqual(0, fut1.get, 0.05)

  def test_onfailure_success(self):
    fut1 = Promise()
    fut2 = Promise.value(1).onfailure(lambda e: fut1.setvalue(e))

    Promise.join([fut2])

    self.assertRaises(TimeoutError, fut1.within(0).get, 0.05)

  def test_onfailure_failure(self):
    e    = ExampleException()
    fut1 = Promise()
    fut2 = Promise.exception(e).onfailure(lambda e: fut1.setvalue(e))

    self.assertEqual(fut1.get(0.5), e)

  def test_onfailure_exception(self):
    def uhoh(e):
      raise ExampleException("uh oh!")

    fut1 = Promise.exception(RuntimeError()).onfailure(uhoh)

    self.assertRaises(RuntimeError, fut1.get, 0.05)


class PromiseMapTests(object):
  """Tests for map/flatmap"""

  def test_flatmap_success(self):
    fut1 = Promise.value(1)
    fut2 = fut1.flatmap(lambda v: Promise.value(v+1))

    self.assertEqual(fut2.get(0.5), 2)

  def test_flatmap_exception(self):
    fut1 = Promise.exception(ExampleException())
    fut2 = fut1.flatmap(lambda v: v+1)

    self.assertRaises(ExampleException, fut2.get, 0.05)

  def test_flatmap_raises(self):
    def uhoh(v):
      raise ExampleException(v)

    fut1 = Promise.value(1)
    fut2 = fut1.flatmap(uhoh)

    self.assertRaises(ExampleException, fut2.get, 0.1)

  def test_flatmap_badfunc(self):
    fut1 = Promise.value(1)
    fut2 = fut1.flatmap(lambda: 1)

    self.assertRaises(TypeError, fut2.get, 0.1)

  def test_flatmap_nonpromise_return_value(self):
    fut1 = Promise.value(1)
    fut2 = fut1.flatmap(lambda v: 1)

    self.assertRaises(MiraiError, fut2.get, 0.1)

  def test_map_success(self):
    fut1 = Promise.value(1)
    fut2 = fut1.map(lambda v: v+1)

    self.assertEqual(fut2.get(0.5), 2)

  def test_map_failure(self):
    fut1 = Promise.exception(ExampleException())
    fut2 = fut1.map(lambda v: v+1)

    self.assertRaises(ExampleException, fut2.get, 0.1)

  def test_map_raises(self):
    def uhoh(v):
      raise ExampleException(v)

    fut1 = Promise.value(1)
    fut2 = fut1.map(uhoh)

    self.assertRaises(ExampleException, fut2.get, 0.1)

  def test_map_badfunc(self):
    fut1 = Promise.value(1)
    fut2 = fut1.map(lambda: 1)
    self.assertRaises(TypeError, fut2.get, 0.1)


class PromiseMiscellaneousTests(object):
  """Other Promise methods..."""

  def test_filter_success(self):
    fut1 = Promise.value(1).filter(lambda v: v != 1)
    self.assertRaises(MiraiError, fut1.get, 0.1)

    fut1 = Promise.value(1).filter(lambda v: v == 1)
    self.assertEqual(fut1.get(0.05), 1)

    fut1 = Promise.exception(MiraiError()).filter(lambda v: v == 1)
    self.assertRaises(MiraiError, fut1.get, 0.1)

  def test_filter_exception(self):
    def uhoh(v):
      raise ExampleException()

    fut1 = Promise.value(1).filter(uhoh)
    self.assertRaises(ExampleException, fut1.get, 0.05)

  def test_filter_badfunc(self):
    fut1 = Promise.value(1).filter(lambda: False)
    self.assertRaises(TypeError, fut1.get, 0.05)

  def test_rescue_success(self):
    fut1 = Promise.value("A")
    fut2 = fut1.rescue(lambda e: "B")

    self.assertEqual(fut2.get(0.5), "A")

  def test_rescue_failure(self):
    fut1 = Promise.exception(ExampleException("A"))
    fut2 = fut1.rescue(lambda e: Promise.value(e.message))

    self.assertEqual(fut2.get(0.5), "A")

  def test_rescue_nonpromise_return_value(self):
    fut1 = Promise.exception(ExampleException("A"))
    fut2 = fut1.rescue(lambda e: "uh oh")

    self.assertRaises(MiraiError, fut2.get, 0.5)

  def test_rescue_raises(self):
    def reraise(e):
      raise e

    fut1 = Promise.exception(ExampleException("A"))
    fut2 = fut1.rescue(reraise)

    self.assertRaises(ExampleException, fut2.get, 0.5)

  def test_rescue_badfunc(self):
    fut1 = Promise.exception(ExampleException("A"))
    fut2 = fut1.rescue(lambda: 1)

    self.assertRaises(TypeError, fut2.get, 0.5)

  def test_transform_success(self):
    fut1 = Promise.value(1)
    fut2 = fut1.transform(lambda p: Promise.exception(ExampleException()))

    self.assertRaises(ExampleException, fut2.get, 0.05)

  def test_transform_bandfunc(self):
    fut1 = Promise.value(1)
    fut2 = fut1.transform(lambda: Promise.value(0))

    self.assertRaises(TypeError, fut2.get, 0.05)

  def test_transform_exception(self):
    def uhoh(fut):
      raise ExampleException()
    fut1 = Promise.value(1)
    fut2 = fut1.transform(uhoh)

    self.assertRaises(ExampleException, fut2.get, 0.05)

  def test_within_success(self):
    fut1 = Promise.value("A")
    fut2 = fut1.within(0.5)

    self.assertEqual(fut2.get(0.5), "A")

  def test_within_failure(self):
    fut1 = Promise()
    fut2 = fut1.within(0)

    self.assertRaises(TimeoutError, fut2.get, 0.05)

  def test_respond_success(self):
    fut1 = Promise()
    Promise.value(1).respond(lambda f: f.proxyto(fut1))

    self.assertEqual(1, fut1.get(0.05))

    fut1 = Promise()
    Promise.exception(MiraiError()).respond(lambda f: f.proxyto(fut1))

    self.assertRaises(MiraiError, fut1.get, 0.05)

  def test_respond_exception(self):
    def uhoh(o):
      raise ExampleException()

    fut1 = Promise.value(1).respond(uhoh)

    self.assertEqual(1, fut1.get(0.05))

  def test_respond_badfunc(self):
    fut1 = Promise.value(1).respond(lambda: 1)
    self.assertEqual(1, fut1.get(0.05))

  def test_unit(self):
    self.assertIsNone(Promise.value(1).unit().get(0.05))
    self.assertRaises(ExampleException, Promise.exception(ExampleException()).unit().get, 0.05)

  def test_update(self):
    # values are propagated
    self.assertEqual(Promise().update(Promise.value(1)).get(0.01), 1)

    # exceptions are propagated
    self.assertRaises(ExampleException, Promise().update(Promise.exception(ExampleException())).get, 0.01)

    # non-Promise argument
    self.assertRaises(MiraiError, Promise().update, 1)

  def test_updateifempty(self):
    # nothing/value
    self.assertEqual(Promise().updateifempty(Promise.value(1)).get(0.01), 1)

    # value/value
    self.assertEqual(Promise().setvalue(2).updateifempty(Promise.value(1)).get(0.01), 2)

    # exception/nothing
    self.assertRaises(MiraiError,
      Promise()
      .setexception(MiraiError())
      .updateifempty(Promise())
      .get,
      0.05
    )

    # nothing/exception
    self.assertRaises(MiraiError,
      Promise()
      .updateifempty(Promise.exception(MiraiError()))
      .get,
      0.05
    )

    # exception/value
    self.assertRaises(MiraiError,
      Promise()
      .setexception(MiraiError())
      .updateifempty(Promise.value(1))
      .get,
      0.05
    )

    self.assertRaises(MiraiError, Promise().updateifempty, 1)

  def test_executor(self):
    old = Promise.executor()
    new = Promise.executor(ThreadPoolExecutor(max_workers=10))
    self.assertEqual(Promise.call(lambda v: v+1, 1).get(0.05), 2)


class PromiseThreadcountTests(object):
  """Tests involving a fixed number of threads"""

  def test_within_many_threads(self):
    # ensure's that within actually works, even when other threads are waiting.
    import time

    Promise.executor(ThreadPoolExecutor(max_workers=20))
    promises = [
      Promise.call(time.sleep, 0.05 if i < 5 else 0.75)
      for i in range(10)
    ]
    promise = (
      Promise.collect(promises)
      .within(0.1)
      .handle(lambda err: "yay")
    )
    self.assertEqual("yay", promise.get(0.5))

  def test_within_few_threads(self):
    # ensure that code doesn't lock up if I create far more threads than I have
    # workers.
    import time

    Promise.executor(ThreadPoolExecutor(max_workers=5))

    def another(i):
      # sleep a little, and a bunch of callbacks
      if i <= 0:
        return Promise.value("done")
      else:
        result = Promise.call(create, i-1)
        for i in range(i):
          result = result.map(lambda i: i)
        time.sleep(0.05)
        return result

    promises = [Promise.call(another, i) for i in range(10)]
    promise  = (
      Promise.collect(promises)
      .within(0.25)
      .handle(lambda err: "yay")
    )

    self.assertEqual("yay", promise.get(0.5))

  def test_select_few_threads(self):
    # this ensures that Promise.select won't cause a threadlock if all workers
    # are busy with the threads it's waiting on.
    Promise.executor(ThreadPoolExecutor(max_workers=2))

    promises = [
      Promise.select([
        Promise.wait(0.5).map(lambda v: v),
        Promise.wait(0.1).map(lambda v: "yay"),
      ])
      .flatmap(lambda (winner, losers): winner)
      for i in range(2)
    ]

    # shouldn't throw a timeout error
    Promise.collect(promises).get(2.5)


class PromiseAlternativeNamesTests(object):
  """Promise methods that are just alternative names for other methods"""

  def test_andthen(self):
    self.assertEqual(
        Promise.value(2).flatmap(lambda v: Promise.value(5)).get(0.5),
        Promise.value(2).andthen(lambda v: Promise.value(5)).get(0.5),
    )

  def test_call(self):
    self.assertEqual(Promise.value(1).get(0.5), Promise.value(1)())
    self.assertRaises(TimeoutError, Promise().get, 0.05)

  def test_ensure(self):
    fut1 = Promise()
    fut2 = Promise.value(2).ensure(lambda: fut1.setvalue(2))

    self.assertEqual(fut1.get(0.5), 2)

    fut1 = Promise()
    fut2 = Promise.exception(ExampleException()).ensure(lambda: fut1.setvalue(2))

    self.assertEqual(fut1.get(0.5), 2)

  def test_ensure_exception(self):
    def uhoh():
      raise ExampleException("uh oh!")

    fut1 = Promise.value(2).ensure(uhoh)

    self.assertEqual(2, fut1.get(0.05))

  def test_foreach(self):
    fut1 = Promise()
    fut2 = Promise.value(1).foreach(lambda v: fut1.setvalue(v))

    self.assertEqual(fut1.get(0.05), 1)

    fut1 = Promise()
    fut2 = Promise.exception(ExampleException()).foreach(lambda v: fut1.setvalue(v))

    self.assertRaises(TimeoutError, fut1.get, 0.05)

  def test_getorelse(self):
    self.assertEqual(Promise().getorelse(0.5), 0.5)
    self.assertEqual(Promise.value(1).getorelse(0.5), 1)
    self.assertEqual(Promise.exception(ExampleException).getorelse(0.5), 0.5)

  def test_handle(self):
    self.assertEqual(
      Promise
        .exception(ExampleException("uh oh"))
        .handle(lambda e: e.message)
        .get(0.05),
      "uh oh",
    )

    self.assertEqual(
      Promise
        .value(1)
        .handle(lambda e: e.message)
        .get(0.05),
      1,
    )

  def test_select_(self):
    self.assertEqual(
      Promise.wait(0.50).flatmap(lambda v: Promise.exception(MiraiError()))
      .select_(
        Promise.wait(0.05).map(lambda v: 2),
        Promise.wait(0.90).map(lambda v: 3),
      ).get(0.40),
      2,
    )


class PromiseMergingTests(object):
  """Methods for combining Promises together"""

  def test_collect_empty(self):
    self.assertEqual(Promise.collect([]).get(0.1), [])

  def test_collect_success(self):
    fut1 = [Promise.value(1), Promise.value(2), Promise.value(3)]
    fut2 = Promise.collect(fut1).within(0.01)

    self.assertEqual(fut2.get(0.5), [1,2,3])

  def test_collect_failure(self):
    fut1 = [Promise.exception(ExampleException()), Promise.value(2), Promise.value(3)]
    fut2 = Promise.collect(fut1)

    self.assertRaises(ExampleException, fut2.get, 0.05)

  def test_join_success(self):
    fut1 = [Promise.wait(0.1).map(lambda v: 0.1), Promise.value(0.1)]
    fut2 = Promise.join(fut1)

    for fut in fut1:
      self.assertEqual(fut.get(0.5), 0.1)

  def test_join_failure(self):
    fut1 = [Promise.value(0), Promise().within(0.05)]
    fut2 = Promise.join(fut1)

    self.assertRaises(TimeoutError, fut2.get, 0.05)

  def test_select(self):
    fut1 = [Promise(), Promise.wait(0.05).map(lambda v: 0.05)]
    resolved, rest = Promise.select(fut1).get(0.1)

    self.assertEqual(len(rest), 1)
    self.assertEqual(resolved.get(0.5), 0.05)
    self.assertFalse(rest[0].isdefined())

  def test_join_(self):
    self.assertEqual(
      Promise.value(1).join_(Promise.value(2)).get(0.05),
      [1,2],
    )

  def test_or_(self):
    self.assertEqual(
      Promise.wait(0.05).map(lambda v: 1).or_(
        Promise.wait(0.25).map(lambda v: 2),
        Promise.wait(0.50).map(lambda v: 3),
      ).get(0.1),
      1,
    )


class FutureTests(
    unittest.TestCase
  ):
  """Tests for read-only Futures"""

  def setUp(self):
    Promise.executor(ThreadPoolExecutor(max_workers=10))

  def test_proxy(self):
    promise = Promise()
    future  = promise.future()

    self.assertRaises(
      TimeoutError,
      future.get,
      timeout=0.01,
    )

    promise.setvalue(1)

    self.assertEqual(1, future.get(timeout=0.01))

  def test_no_set(self):
    future = Promise().future()
    self.assertRaises(AttributeError, future.setvalue, 1)
    self.assertRaises(AttributeError, future.setexception, ExampleException())


class PromiseTests(
    PromiseConstructorTests,
    PromiseBasicTests,
    PromiseCallbackTests,
    PromiseMapTests,
    PromiseMiscellaneousTests,
    PromiseThreadcountTests,
    PromiseAlternativeNamesTests,
    PromiseMergingTests,

    unittest.TestCase,
  ):

  def setUp(self):
    Promise.executor(ThreadPoolExecutor(max_workers=10))


if __name__ == '__main__':
  unittest.main()



from concurrent.futures import TimeoutError
from .futures import Promise, Future
from .exceptions import AlreadyResolvedError, MiraiError
from .pool import UnboundedThreadPoolExecutor
from ._version import __version__



from setuptools import setup, find_packages

def version(name):
  import os
  folder   = os.path.split(__file__)[0]
  mirai    = os.path.join(folder, name)
  _version = os.path.join(mirai, "_version.py")
  env      = {}
  execfile(_version, env)
  return env['__version__']


if __name__ == '__main__':
  setup(
      name              = 'mirai',
      version           = version('mirai'),
      author            = 'Daniel Duckworth',
      author_email      = 'duckworthd@gmail.com',
      description       = "Twitter Futures in Python",
      license           = 'BSD',
      keywords          = 'futures concurrent finagle twitter',
      url               = 'http://github.com/duckworthd/mirai',
      packages          = find_packages(),
      classifiers       = [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
      ],
      setup_requires    = [
        "setuptools>=3.4.4",
      ],
      install_requires  = [
        "futures>=2.1.6",
        "joblib>=0.7.1",
      ],
      tests_require     = [
        "nose>=1.3.1",
      ],
      test_suite = "nose.collector",
  )



from concurrent import futures
from concurrent.futures import TimeoutError
import logging
import threading
import time
import traceback

from .exceptions import MiraiError, SafeFunction, AlreadyResolvedError
from .utils import proxyto

# Future methods:
#   cancel()
#   cancelled()
#   running()
#   done()
#   result(timeout=None)
#   exception(timeout=None)
#   add_done_callback(fn)
#   set_result(result)
#   set_exception(exception)
#
# futures.wait(fs, timeout=None, return_when=ALL_COMPLETED)
#   (alt: FIRST_COMPLETED, FIRST_EXCEPTION)
# futures.as_completed(fs, timeout=None)
#
# Executor methods:
#   submit(fn *args, **kwargs)
#   map(fn, iterables)
#   shutdown()


class Promise(object):
  """
  A `Promise` encapsulates the result of an asynchronous computation. Think of
  it as a single-use mailbox -- you receive a promise which will later contain
  a message.::

    import requests
    from mirai import Promise

    def async_request(method, url, *args, **kwargs):
      "fetch a url asynchronously using `requests`"

      # construct a promise to fill later
      promise = Promise()

      def sync_request():
        "fetches synchronously & propagates exceptions"
        try:
          response = requests.request(method, url, *args, **kwargs)
          promise.setvalue(response)
        except Exception as e:
          promise.setexception(e)

      # start asynchronous computation
      Promise.call(sync_request)

      # return read-only version of promise
      return promise.future()

  Parameters
  ----------
  future : concurrent.futures.Future
      Future this promise wraps.
  """

  EXECUTOR = futures.ThreadPoolExecutor(max_workers=10)

  __slots__ = ['_future', '_lock']

  def __init__(self, future=None):
    self._future = future or futures.Future()
    self._lock   = threading.Lock()

  def andthen(self, fn):
    """
    Apply a function with a single argument: the value this Promise resolves to.
    The function must return another future.  If this Promise fails, `fn` will
    not be called. Same as as `Promise.flatmap`.

    Parameters
    ----------
    fn : (value,) -> Promise
        Function to apply. Takes 1 positional argument. Must return a Promise.

    Returns
    -------
    result : Future
        Promise `fn` will return.
    """
    return self.flatmap(fn)

  def __call__(self, timeout=None):
    """
    Retrieve value of Promise; block until it's ready or `timeout` seconds
    have passed. If `timeout` seconds pass, then a `TimeoutError` will
    be raised. If this Promise failed, the set exception will be raised.

    Parameters
    ----------
    timeout : number or None
        Number of seconds to wait until raising a `TimeoutError`. If `None`,
        then wait indefinitely.

    Returns
    -------
    result : object
        Contents of this future if it resolved successfully.

    Raises
    ------
    Exception
        Set exception if this future failed.
    """
    return self.get(timeout)

  def ensure(self, fn):
    """
    Ensure that no-argument function `fn` is called when this Promise resolves,
    regardless of whether or not it completes successfuly.

    Parameters
    ----------
    fn : (,) -> None
        function to apply upon Promise completion. takes no arguments. Return
        value ignored.

    Returns
    -------
    self : Future
    """
    return self.respond(lambda fut: Promise.eval(fn).get())

  def filter(self, fn):
    """
    Construct a new Promise that fails if `fn` doesn't evaluate truthily when
    given `self.get()` as its only argument. If `fn` evaluates falsily, then
    the resulting Promise fails with a `MiraiError`.

    Parameters
    ----------
    fn : (value,) -> bool
        function used to check `self.get()`. Must return a boolean-like value.

    Returns
    -------
    result : Future
        Future whose contents are the contents of this Promise if `fn` evaluates
        truth on this Promise's contents.
    """
    return self.flatmap(lambda v:
      Promise.value(v) if Promise.eval(fn, v).get()
      else Promise.exception(MiraiError(u"Value {} was filtered out".format(v)))
    )

  def flatmap(self, fn):
    """
    Apply a function with a single argument: the value this Promise resolves to.
    The function must return another future.  If this Promise fails, `fn` will
    not be called.

    Parameters
    ----------
    fn : (value,) -> Promise
        Function to apply. Takes 1 positional argument. Must return a Promise.

    Returns
    -------
    result : Future
        Future containing return result of `fn`.
    """
    def flatmap(fut):
      assert fut.isdefined(), (
        u"Internal callback in Promise.flatmap was called before the Promise "
        u"it's responsible for finished. How did you make that happen? "
      )

      try:
        result = Promise.eval(fn, fut.get()).get()
      except Exception as e:
        return Promise.exception(e)
      else:
        if not isinstance(result, Promise):
          raise MiraiError(u"Functions chained with Promise.flatmap must return Promises!")
        return result

    return self.transform(flatmap)

  def foreach(self, fn):
    """
    Apply a function if this Promise resolves successfully. The function
    receives the contents of this Promise as its only argument.

    Parameters
    ----------
    fn : (value,) -> None
        Function to apply to this Promise's contents. Return value ignored.

    Returns
    -------
    self : Promise
    """
    return self.onsuccess(fn)

  def future(self):
    """
    Retrieve a `Future` encapsulating this promise. A `Future` is a read-only
    version of the exact same thing.

    Returns
    -------
    future : Future
        Future encapsulating this Promise.
    """
    return Future(self)

  def get(self, timeout=None):
    """
    Retrieve value of Promise; block until it's ready or `timeout` seconds
    have passed. If `timeout` seconds pass, then a `TimeoutError` will
    be raised. If this Promise failed, the set exception will be raised.

    Parameters
    ----------
    timeout : number or None
        Number of seconds to wait until raising a `TimeoutError`. If `None`,
        then wait indefinitely.

    Returns
    -------
    result : anything
        Contents of this future if it resolved successfully.

    Raises
    ------
    Exception
        Set exception if this future failed.
    """
    return self._future.result(timeout)

  def getorelse(self, default, timeout=None):
    """
    Like `Promise.get`, but instead of raising an exception when this Promise
    fails, returns a default value.

    Parameters
    ----------
    default : anything
        default value to return in case of timeout or exception.
    timeout : None or float
        time to wait before returning default value if this promise is unresolved.

    Returns
    -------
    result : anything
        value this Promise resolves to, if it resolves successfully, else
        `default`.
    """
    try:
      if timeout is None:
        timeout = 0
      return self.get(timeout=timeout)
    except Exception as e:
      return default

  def handle(self, fn):
    """
    If this Promise fails, call `fn` on the ensuing exception to obtain a
    successful value.

    Parameters
    ----------
    fn : (exception,) -> anything
        Function applied to recover from a failed exception. Its return value
        will be the value of the resulting Promise.

    Returns
    -------
    result : Future
        Resulting Future returned by applying `fn` to the exception, then
        setting the return value to `result`'s value. If this Promise is
        already successful, its value is propagated onto `result`.
    """
    return self.rescue(lambda e: Promise.eval(fn, e))

  def isdefined(self):
    """
    Return `True` if this Promise has already been resolved, successfully or
    unsuccessfully.

    Returns
    -------
    result : bool
    """
    return self._future.done()

  def isfailure(self):
    """
    Return `True` if this Promise failed, `False` if it succeeded, and `None` if
    it's not yet resolved.

    Returns
    -------
    result : bool
    """
    v = self.issuccess()
    if v is None : return None
    else         : return not v

  def issuccess(self):
    """
    Return `True` if this Promise succeeded, `False` if it failed, and `None` if
    it's not yet resolved.

    Returns
    -------
    result : bool
    """
    if not self.isdefined():
      return None
    else:
      try:
        self.get()
        return True
      except Exception as e:
        return False

  def join_(self, *others):
    """
    Combine values of this Promise and 1 or more other Promises into a list.
    Results are in the same order `[self] + others` is in.

    Parameters
    ----------
    others : 1 or more Promises
        Promises to combine with this Promise.

    Returns
    -------
    result : Future
        Future resolving to a list of containing the values of this Promise and
        all other Promises. If any Promise fails, `result` holds the exception in
        the one which fails soonest.
    """
    return Promise.collect([self] + list(others))

  def map(self, fn):
    """
    Transform this Promise by applying a function to its value. If this Promise
    contains an exception, `fn` is not applied.

    Parameters
    ----------
    fn : (value,) -> anything
        Function to apply to this Promise's value on completion.

    Returns
    -------
    result : Future
        Future containing `fn` applied to this Promise's value. If this Promise
        fails, the exception is propagated.
    """
    return self.flatmap(lambda v: Promise.eval(fn, v))

  def onfailure(self, fn):
    """
    Apply a callback if this Promise fails. Callbacks can be added after this
    Promise has resolved. If `fn` throws an exception, a warning is printed via
    `logging`.

    Parameters
    ----------
    fn : (exception,) -> None
        Function to call upon failure. Its only argument is the exception set
        to this Promise. If this future succeeds, `fn` will not be called.

    Returns
    -------
    self : Promise
    """
    def onfailure(fut):
      assert fut.isdefined(), (
        u"Internal callback in Promise.onfailure was called before the Promise "
        u"it's responsible for finished. How did you make that happen? "
      )
      try:
        v = fut.get()
      except Exception as e:
        Promise.eval(fn, e).get()

    return self.respond(onfailure)

  def onsuccess(self, fn):
    """
    Apply a callback if this Promise succeeds. Callbacks can be added after this
    Promise has resolved. If `fn` throws an exception, a warning is printed via
    `logging`.

    Parameters
    ----------
    fn : (value,) -> None
        Function to call upon success. Its only argument is the value set
        to this Promise. If this future fails, `fn` will not be called.

    Returns
    -------
    self : Promise
    """
    def onsuccess(fut):
      assert fut.isdefined(), (
        u"Internal callback in Promise.onsuccess was called before the Promise "
        u"it's responsible for finished. How did you make that happen? "
      )
      try:
        v = fut.get()
      except Exception as e:
        pass
      else:
        Promise.eval(fn, v).get()

    return self.respond(onsuccess)

  def or_(self, *others):
    """
    Return the first Promise that finishes among this Promise and one or more
    other Promises.

    Parameters
    ----------
    others : one or more Promises
        Other futures to consider.

    Returns
    -------
    result : Future
        First future that is resolved, successfully or otherwise.
    """
    return Promise.select((self,) + others).flatmap(lambda (fut, futs): fut)

  def proxyto(self, other):
    """
    Copy the state of this Promise to another.

    Parameters
    ----------
    other : Promise
        Another Promise to copy the state of this Promise to, upon completion.

    Returns
    -------
    self : Promise

    Raises
    ------
    MiraiError
        if `other` isn't a Promise instance
    """
    if not isinstance(other, Promise):
      raise MiraiError(
        "Attempting to proxy this Promise onto a non-Promise object. You can "
        "only copy the state of one Promise to another. "
      )
    self.onsuccess(other.setvalue).onfailure(other.setexception)
    return self

  def rescue(self, fn):
    """
    If this Promise fails, call `fn` on the ensuing exception to recover another
    (potentially successful) Promise. Similar to `Promise.handle`, but must
    return a Promise (rather than a value).

    Parameters
    ----------
    fn : (exception,) -> Promise
        Function applied to recover from a failed exception. Must return a Promise.

    Returns
    -------
    result : Future
        Resulting Future returned by apply `fn` to the exception this Promise
        contains. If this Promise is successful, its value is propagated onto
        `result`.
    """
    def rescue(fut):
      try:
        assert fut.isdefined(), (
          u"Internal callback in Promise.rescue was called before the Promise "
          u"it's responsible for finished. How did you make that happen? "
        )

        try:
          v = fut.get()
        except Exception as e:
          # Calling fn via Promise.eval(...).get() guarantees that any
          # exceptions thrown will be wrapped in a SafeFunction, thus
          # preserving the traceback context.
          result = Promise.eval(fn, e).get()
          if not isinstance(result, Promise):
            raise MiraiError(u"Functions chained with Promise.rescue must return Futures!")
          return result
        else:
          return Promise.value(v)

      except Exception as e:
        return Promise.exception(e)

    return self.transform(rescue)

  def respond(self, fn):
    """
    Apply a function to this Promise when it's resolved. If `fn` raises an
    exception a warning will be printed via `logging`, but no action will be
    taken.

    Parameters
    ----------
    fn : (future,) -> None
        Function to apply to this Promise upon completion. Return value is ignored

    Returns
    -------
    self : Promise
    """
    def done_callback(fut):
      try:
        Promise.eval(fn, self).get()
      except Exception as e:
        log = logging.getLogger(__name__ + ".Promise")
        log.warning(u"Function {} in Promise.respond threw an exception!".format(fn))
        log.warning(e)
        log.warning(traceback.format_exc())

    self._future.add_done_callback(done_callback)

    return self

  def select_(self, *others):
    """
    Return the first Promise that finishes among this Promise and one or more
    other Promises.

    Parameters
    ----------
    others : one or more Promises
        Other futures to consider.

    Returns
    -------
    result : Future
        First future that is resolved, successfully or otherwise.
    """
    return self.or_(*others)

  def setexception(self, e):
    """
    Set the state of this Promise as failed with a given Exception. State can
    only be set once; once a Promise is defined, it cannot be redefined. This
    operation is thread (but not process) safe.

    Parameters
    ----------
    e : Exception

    Returns
    -------
    self : Promise

    Raises
    ------
    AlreadyResolvedError
        if this Promise's value is already set
    """
    with self._lock:
      if self.isdefined():
        raise AlreadyResolvedError(u"Promise is already resolved; you cannot set its status again.")
      else:
        self._future.set_exception(e)
        return self

  def setvalue(self, val):
    """
    Set the state of this Promise as successful with a given value. State can
    only be set once; once a Promise is defined, it cannot be redefined. This
    operation is thread (but not process) safe.

    Parameters
    ----------
    val : value

    Returns
    -------
    self : Promise

    Raises
    ------
    AlreadyResolvedError
        if this Promise's value is already set
    """
    with self._lock:
      if self.isdefined():
        raise AlreadyResolvedError(u"Promise is already resolved; you cannot set its status again.")
      else:
        self._future.set_result(val)
        return self

  def transform(self, fn):
    """
    Apply a function with a single argument (this Promise) after resolving.
    The function must return another future.

    Parameters
    ----------
    fn : (future,) -> Promise
        Function to apply. Takes 1 positional argument. Must return a Promise.

    Returns
    -------
    result : Future
        Future containing return result of `fn`.
    """
    p = Promise()
    def transform(fut):
      assert fut.isdefined(), (
        u"Internal callback in Promise.transform was called before the Promise "
        u"it's responsible for finished. How did you make that happen? "
      )

      try:
        q = Promise.eval(fn, fut).get()
      except Exception as e:
        p.setexception(e)
      else:
        if not isinstance(q, Promise):
          raise MiraiError(u"Functions chained with Promise.transform must return Futures!")
        p.update(q)

    self.respond(transform)

    return p

  def unit(self):
    """
    Convert this Promise to another that disregards its result.

    Returns
    -------
    result : Future
        Promise with a value of `None` if this Promise succeeds. If this Promise
        fails, the exception is propagated.
    """
    return self.map(lambda v: None)

  def update(self, other):
    """
    Populate this Promise with the contents of another.

    Parameters
    ----------
    other : Promise
        Promise to copy

    Returns
    -------
    self : Promise

    Raises
    ------
    MiraiError
        if `other` isn't a Promise
    """
    if not isinstance(other, Promise):
      raise MiraiError(
        "`Promise.update`'s argument isn't a Promise. You can't copy the state "
        "of this Promise onto a non-Promise object."
      )

    other.proxyto(self)
    return self

  def updateifempty(self, other):
    """
    Like `Promise.update`, but update only if this Promise isn't already defined.

    Parameters
    ----------
    other : Promise
        Promise to copy, if necessary.

    Returns
    -------
    self : Promise

    Raises
    ------
    MiraiError
        if `other` isn't a Promise
    """
    if not isinstance(other, Promise):
      raise MiraiError(
        "`Promise.updateifempty`'s argument isn't a Promise. You can't copy the "
        "state of this Promise onto a non-Promise object."
      )

    def setvalue(v):
      try:
        self.setvalue(v)
      except AlreadyResolvedError as e:
        pass
      except Exception as e:
        self.setexception(e)

    def setexception(e):
      try:
        self.setexception(e)
      except AlreadyResolvedError as e_:
        pass
      except Exception as e_:
        self.setexception(e_)

    other.onsuccess(setvalue).onfailure(setexception)
    return self

  def within(self, duration):
    """
    Return a Promise whose state is guaranteed to be resolved within `duration`
    seconds. If this Promise completes before `duration` seconds expire, it will
    contain this Promise's contents. If this Promise is not resolved by then, the
    resulting Promise will fail with a `TimeoutError`.

    Parameters
    ----------
    duration : number
        Number of seconds to wait before resolving a `TimeoutError`

    Returns
    -------
    result : Promise
        Promise guaranteed to resolve in `duration` seconds.
    """
    e = TimeoutError(u"Promise did not finish in {} seconds".format(duration))
    return self.or_(Promise.wait(duration).flatmap(lambda v: Promise.exception(e)))

  # CONSTRUCTORS
  @classmethod
  def exception(cls, exc):
    """
    Construct a Promise that has already failed with a given exception.

    Parameters
    ----------
    exc : Exception
        Exception to fail new Promise with

    Returns
    -------
    result : Future
        New Promise that has already failed with the given exception.
    """
    f = cls()
    f.setexception(exc)
    return f.future()

  @classmethod
  def value(cls, val):
    """
    Construct a Promise that is already resolved successfully to a value.

    Parameters
    ----------
    val : anything
        Value to resolve new Promise to.

    Returns
    -------
    result : Future
        Future containing `val` as its value.
    """
    f = cls()
    f.setvalue(val)
    return f.future()

  @classmethod
  def wait(cls, duration):
    """
    Construct a Promise that succeeds in `duration` seconds with value `None`.

    Parameters
    ----------
    duration : number
        Number of seconds to wait before resolving a `TimeoutError`

    Returns
    -------
    result : Future
        Promise that will resolve in `duration` seconds with value `None`.
    """
    def wait():
      try:
        time.sleep(duration)
        return None
      except Exception as e:
        raise e

    return Promise.call(wait).future()

  @classmethod
  def collect(cls, fs):
    """
    Convert a sequence of Promises into a Promise containing a sequence of
    values, one per Promise in `fs`. The resulting Promise resolves once all
    Promises in `fs` resolve successsfully or upon the first failure. In the
    latter case, the failing Promise's exception is propagated.

    Parameters
    ----------
    fs : [Promise]
        List of Promises to merge.

    Returns
    -------
    result : Future
        Future containing values of all Futures in `fs`. If any Future in `fs`
        fails, `result` fails with the same exception.
    """
    if len(fs) == 0:
      # Promise below will never fulfill if there are no fs
      return Promise.value([])
    else:
      lock             = threading.RLock()
      results          = [None for i in range(len(fs))]
      result_container = Promise()

      # for some odd reason, this variable isn't accessible in the `body`
      # closure below if it's just a naive integer. Boxing it in a list somehow
      # makes it work(?)
      count            = [len(fs)]

      for i in range(len(fs)):

        # unless i is "captured" by a function call, all elements will use i ==
        # len(fs)-1.
        def body(i):

          # immediately set result_container's value to the thrown exception
          def onfailure(e):
            (
              result_container
              .updateifempty(Promise.exception(e))
            )

          # fill in one more value in the results array. if this is the last
          # future to return successfully, set the value of the result
          # container.
          def onsuccess(v):
            with lock:
              results[i] = v
              count[0]  -= 1
              if count[0] == 0:
                (
                  result_container
                  .updateifempty(Promise.value(results))
                )

          # attach success and failure callbacks to this future
          (
            fs[i]
            .onsuccess(onsuccess)
            .onfailure(onfailure)
          )

        body(i)

      return result_container

  @classmethod
  def join(cls, fs):
    """
    Construct a Promise that resolves when all Promises in `fs` have resolved. If
    any Promise in `fs` fails, the error is propagated into the resulting
    Promise.

    Parameters
    ----------
    fs : [Promise]
        List of Promises to merge.

    Returns
    -------
    result : Future
        Future containing None if all Futures in `fs` succeed, or the exception
        of the first failing Future in `fs`.
    """
    return cls.collect(fs).map(lambda v: None)

  @classmethod
  def select(cls, fs):
    """
    Return a Promise containing a tuple of 2 elements. The first is the first
    Promise in `fs` to resolve; the second is all remaining Promises that may or
    may not be resolved yet. The resolved Promise is not guaranteed to have
    completed successfully.

    Parameters
    ----------
    fs : [Promise]
        List of Promises to merge.

    Returns
    -------
    result : Future
        Future containing the first Future in `fs` to finish and all remaining
        (potentially) unresolved Futures as a tuple of 2 elements for its value.
    """
    p = Promise()
    if len(fs) == 0:
      raise MiraiError('Promise.select requires at least one future')
    else:
      for f in fs:
        # see comment in Promise.collect regarding loops and closures
        def body(f):
          # all futures, other than the one executing this callback
          not_me = [g for g in fs if g != f]
          f.respond(lambda f: \
            p.updateifempty(Promise.value( (f, not_me) ))
          )
        body(f)
    return p

  @classmethod
  def eval(cls, fn, *args, **kwargs):
    """
    Call a function (synchronously) and return a Promise with its result. If an
    exception is thrown inside `fn`, a new exception type will be constructed
    inheriting both from `MiraiError` and the exception's original type. The
    new exception is the same the original, except that it also contains a
    `context` string detailing the stack at the time the exception was thrown.

    Parameters
    ----------
    fn : function
        Function to be called
    *args : arguments
    **kwargs : keyword arguments

    Returns
    -------
    result : Future
        Future containing the result of `fn(*args, **kwargs)` as its value or
        the exception thrown as its exception.
    """
    if not isinstance(fn, SafeFunction):
      fn = SafeFunction(fn)
    try:
      v = fn(*args, **kwargs)
    except Exception as e:
      return cls.exception(e).future()
    else:
      return cls.value(v).future()

  @classmethod
  def eval(cls, fn, *args, **kwargs):
    """
    Call a function (synchronously) and return a Promise with its result. If an
    exception is thrown inside `fn`, a new exception type will be constructed
    inheriting both from `MiraiError` and the exception's original type. The
    new exception is the same the original, except that it also contains a
    `context` attribute detailing the stack at the time the exception was
    thrown.

    Parameters
    ----------
    fn : function
        Function to be called
    *args : arguments
    **kwargs : keyword arguments

    Returns
    -------
    result : Future
        Future containing the result of `fn(*args, **kwargs)` as its value or
        the exception thrown as its exception.
    """
    fn = SafeFunction(fn)
    try:
      v = fn(*args, **kwargs)
    except Exception as e:
      return cls.exception(e)
    else:
      return cls.value(v)

  @classmethod
  def call(cls, fn, *args, **kwargs):
    """
    Call a function asynchronously and return a Promise with its result. If an
    exception is thrown inside `fn`, a new exception type will be constructed
    inheriting both from `MiraiError` and the exception's original type. The
    new exception is the same the original, except that it also contains a
    `context` attribute detailing the stack at the time the exception was
    thrown.

    Parameters
    ----------
    fn : function
        Function to be called
    *args : arguments
    **kwargs : keyword arguments

    Returns
    -------
    result : Future
        Future containing the result of `fn(*args, **kwargs)` as its value or
        the exception thrown as its exception.
    """
    return cls(cls.EXECUTOR.submit(SafeFunction(fn), *args, **kwargs)).future()

  @classmethod
  def executor(cls, executor=None, wait=True):
    """
    Set/Get the EXECUTOR Promise uses. If setting, the current executor is
    first shut down.

    Parameters
    ----------
    executor : concurrent.futures.Executor or None
        If None, retrieve the current executor, otherwise, shutdown the current
        Executor object and replace it with this argument.
    wait : bool, optional
        Whether or not to block this thread until all workers are shut down
        cleanly.

    Returns
    -------
    executor : Executor
        Current executor
    """
    if executor is None:
      return cls.EXECUTOR
    else:
      if cls.EXECUTOR is not None:
        cls.EXECUTOR.shutdown(wait=wait)
      cls.EXECUTOR = executor
      return cls.EXECUTOR


class Future(Promise):
  """Read-only version of a Promise."""

  def __init__(self, promise):
    allowed_specials = [
      '__str__',
      '__unicode__',
      '__repr__',
      '__call__',
    ]
    proxyto(self, promise, allowed_specials)

  def setvalue(self, val):
    raise AttributeError(u"Futures are read only; Promises are writable")

  def setexception(self, val):
    raise AttributeError(u"Futures are read only; Promises are writable")

